#!/bin/sh
#
# written by Thorsten Glebe, Jan. 1999
# Max-Planck-Institut of Nuclear Physics
# Heidelberg / Germany
#
# 1. argument: directory where to copy all the stuff (optional)
#
# per default the binaries are copied to $LIBROOT/$HBBINTYPE/bin
#

DIR=`dirname $0`
source $DIR/util.sh

if [ $# -gt 1 ]; then
    echo "usage: $PROG <directory>"
    exit 1
fi

if [ $# -eq 1 ]; then
  BIN_DIR=${1}/${HBBINTYPE}/bin
else
  BIN_DIR=${LIBROOT}/${HBBINTYPE}/bin
fi

PACK_FILE=${LIBROOT}/PACKAGES

#
# remove old BIN_DIR
#
if [ -d "$BIN_DIR" ]; then
  echo "Directory $BIN_DIR already exists, removing it!"
  rm -rf $BIN_DIR
  check_status "error removing $BIN_DIR"
fi

#
# create BIN_DIR, if not yet existing
#
if [ ! -d "$BIN_DIR" ]; then
  echo "creating $BIN_DIR..."
  mkdir -p $BIN_DIR
  check_status "error creating $BIN_DIR"
fi

#
# collect executables 
#
for i in `cat $PACK_FILE | ${AWK} '!/^[ ]*#.*$/ && $2 ~ /\/[0-9]+-[0-9]+/ {gsub("/",".",$2); print $1"+"substr($2,2)}'`; do
    NAME=`echo $i | cut -d+ -f1`
    VERS=`echo $i | cut -d+ -f2`
    BINNAME=$ITRROOT/packages/$NAME/$NAME$VERS/bin.$HBBINTYPE
    if [ -d "$BINNAME" ]; then
       mv ${BINNAME}/* $BIN_DIR 2>>$LIBROOT/ERRORS
       # ignore errors
       # check_status "error copying from $BINNAME to $BIN_DIR"
    fi
done

exit 0
