#! /bin/bash
#
# written by Rainer Wanke, June 1998, based on itr_collect.sh
# written by Thorsten Glebe, November 1997
# modified by Thorsten Glebe, February 1999
#
# Max-Planck-Institut of Nuclear Physics
# Heidelberg / Germany
#
# 1. argument: directory where to copy all the stuff (optional)
#

DIR=`dirname $0`
source $DIR/util.sh

if [ $# -gt 1 ]; then
    echo "usage: $PROG <directory>"
    exit 1
fi
if [ $# -eq 1 ]; then
    LIBROOT=$1
fi

PACK_FILE=$LIBROOT/PACKAGES


#
#  remove include-directory, if existent
#
if [ ! -d "$LIBROOT" ]; then
  echo "creating $LIBROOT..."
  mkdir $LIBROOT
  check_status "error creating $LIBROOT"
fi
if [ -d $LIBROOT/include ]; then
  echo "$LIBROOT/include exists, removing it!"
  rm -rf $LIBROOT/include
  check_status "error removing $LIBROOT/include"
fi

#
#  create new include-directory
#
echo "creating $LIBROOT/include..."
mkdir $LIBROOT/include
check_status "error creating $LIBROOT/include"

#
# collect header files and copy to itrlib-area
#
for i in `cat $PACK_FILE | ${AWK} '!/^[ ]*#.*$/ && $2 ~ /\/[0-9]+-[0-9]+/ {gsub("/",".",$2); print $1"+"substr($2,2)}'`; do
    NAME=`echo $i | cut -d+ -f1`
    VERS=`echo $i | cut -d+ -f2`
    SRCNAME=$ITRROOT/packages/$NAME/$NAME$VERS/src
    if [ -d "$SRCNAME" ]; then
      find $SRCNAME \( -name *\\.h -o -name *\\.hh -o -name CVStag.$NAME -o -name silmainh.c -o -name silfrlibh.c \) -print -exec cp -p {} $LIBROOT/include \; 2>>$LIBROOT/ERRORS
    else
      echo "$SRCNAME does not exist..."
    fi
done

#
# create ITR.h
#
cd $LIBROOT/include
echo Create ITR.h...
if [ -f ITR.h ]; then
  rm ITR.h
fi
echo "#ifndef __ITRLIB_H_"  > VDS.h
echo "#define __ITRLIB_H_" >> VDS.h
echo "/* Automatic created ITR.h - VDS Project global header file." >> VDS.h
echo " * If possible, use the package-specific header files." >> ITR.h
echo " */" >> ITR.h
for i in *; do
  if [ "$i" != "ITR.h" ]; then
    echo '#include "'$i'"' >> ITR.h
  fi
done
echo "#endif" >> ITR.h
cd -

exit 0

