#include "ITR_marvin.h"



int ITR_marvin_subtractSpline(int *istat, float *y, float CUT, int *outlier)
/*-----------------------------------------------------------------------------
 * 1st order spline baseline subtraction
 * =====================================
 *
 * Author: Michael Schmelling
 *---------------------------------------------------------------------------*/
{

  /* define the spline by the first points on subsequent segments */ 

  static int   ix[NDIM] = {0, 30, 58, 92, NCHAN}; 
  /*------------------------------------------*/

  union {float f; int i;} cut;
  union {float f; int i;} val;

  int    i, j, k,ilow, ihig, ndrop;
  float  t1, t2;

  static int   first=1;            /* initialization flag */
  static int   interval[NCHAN];
  static float a[NCHAN], b[NCHAN], saav[NCHAN], sabv[NCHAN], sbbv[NCHAN];
  static float z[NDIM], sy[NDIM];
  static float saas[NINT], sabs[NINT], sbbs[NINT]; 
  static float saa[NINT], sab[NINT], sbb[NINT];
  /*-----------------------------------------------------*/

  /* on first call setup the spline fit */

  if(first) {
    first = 0;

    for(j=0; j<NINT; ++j) {
      saa[j] = sab[j] = sbb[j] = 0;
      for(i=ix[j]; i<ix[j+1]; ++i) {
        a[i] = ((float)(ix[j+1] - i) - 0.5)/(float)(ix[j+1] - ix[j]);
        b[i] = ((float)(i  -  ix[j]) + 0.5)/(float)(ix[j+1] - ix[j]);  
        saav[i]  = a[i]*a[i];
        sabv[i]  = a[i]*b[i];
        sbbv[i]  = b[i]*b[i];
        saas[j] += saav[i];
        sabs[j] += sabv[i];
        sbbs[j] += sbbv[i];
        interval[i] = j;
      }   
    }
  }

  /* first iteration */

  for(j=0; j<NDIM; ++j) sy[j] = 0;

  for(j=0; j<NINT; ++j) {
    saa[j] = saas[j]; 
    sab[j] = sabs[j]; 
    sbb[j] = sbbs[j];
    ilow = ix[j]; 
    ihig = ix[j+1];
    for(t1=t2=0, i=ilow; i<ihig; ++i) {
      t1 += y[i]*a[i]; 
      t2 += y[i]*b[i];
    }
    sy[j]   += t1;
    sy[j+1] += t2;
  }

  /* subtract contributions from dead channels */
 
  for(k=1; k<istat[0]; k++) {
    i = istat[k];
    j = interval[i];
    sy[j]   -= y[i]*a[i];
    sy[j+1] -= y[i]*b[i];
    saa[j] -= saav[i];
    sab[j] -= sabv[i];
    sbb[j] -= sbbv[i];
  }

  /* save the sy vector */

  for(i=0; i<NDIM; i++) z[i] = sy[i];

  ITR_marvin_solveSpline(saa, sab, sbb, sy);

  /* subtract the baseline */
 
  for(j=0; j<NINT; ++j) {
    ilow = ix[j]; 
    ihig = ix[j+1];
    t1   = sy[j];
    t2   = sy[j+1];
    for(i=ilow; i<ihig; ++i) {
      y[i] -= t1*a[i] + t2*b[i];
    }
  }

  /* zero dead channels */

  for(k=1; k<istat[0]; k++) y[istat[k]]=0;

  /* outlier removal if CUT!=0 */

  cut.f = CUT;
  if( ! cut.i ) return 0; 

  ndrop = 0;

  for(i=0; i<NCHAN; i++) {
    val.f = y[i]; 
    if( (val.i & MASK) > cut.i ) {
      outlier[ndrop++] = i;
      j  = interval[i];
      t1 = val.f + sy[j]*a[i] + sy[j+1]*b[i];
      z[j]   -= t1*a[i];
      z[j+1] -= t1*b[i];
      saa[j] -= saav[i];
      sab[j] -= sabv[i];
      sbb[j] -= sbbv[i];
    }
  }

  if(!ndrop) return 0;

  ITR_marvin_solveSpline(saa, sab, sbb, z);

  /* fix the baseline */
 
  for(j=0; j<NINT; ++j) {
    ilow = ix[j]; 
    ihig = ix[j+1];
    t1   = -sy[j]   + z[j];
    t2   = -sy[j+1] + z[j+1];
    for(i=ilow; i<ihig; ++i) {
      y[i] -= t1*a[i] + t2*b[i];
    }
  }

  /* zero dead channels */

  for(k=1; k<istat[0]; k++) y[istat[k]]=0;

  return ndrop;
}


void ITR_marvin_solveSpline(float *saa, float *sab, float *sbb, float *sy)
{/*===================================================================
  * Auxiliary function to calculate spline parameters from cumulants
  *==================================================================*/

  int   i, k;
  float dd[NMAT];
  /*------------*/

  /* fill the spline fit matrix */

  k=0;
  dd[k++] = saa[0];
  dd[k++] = sab[0];
  for(i=1; i<NINT; i++) {
    dd[k++] = sab[i-1];
    dd[k++] = saa[i] + sbb[i-1];
    dd[k++] = sab[i];
  }
  dd[k++] = sab[NINT-1];
  dd[k++] = sbb[NINT-1];

  /* solve for the coefficients */
 
  i=1; k=-1;
  do { k+=3;
       dd[k]   /= dd[k-2]; 
       dd[k+1] -= dd[k] * dd[k-1];
       sy[i]   -= dd[k] * sy[i-1]; 
  } while(++i<NDIM);
  sy[--i] /= dd[++k];
  do { k-=3; --i;
       sy[i]   -= dd[k+1]*sy[i+1];
       sy[i]   /= dd[k]; 
  } while(i);

  return;
}

 
