#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "iol.h"
#include "aframelib.h"
#include "grr.h"

#define IPRINT_OFF 0
#define IPRINT_ON  1

#define IOL_NVERB_PROTECTED 1

/* Global Variables */
static IOL_STRUCT iol_struct={
  IOL_NAME, IOL_IOL_NVERBS,
  {"version"              ,  IOL_VERSION,      IOL_TYPE_STRING},
  {"debug mode"           ,  0          , " ", IOL_TYPE_FLAG},
  {"match verb exactly"   ,  1          , " ", IOL_TYPE_FLAG},
  {"protect version"      ,  1          , " ", IOL_TYPE_FLAG},
  {"report unknown lines" ,  1          , " ", IOL_TYPE_FLAG}
};

static IOL_LIST iol_list = NULL;

static char iol_protected_verb[IOL_NVERB_PROTECTED][IOL_NLEN]={"version"};

static int paraminit_done    = 0; /* Flag that param_init was done */

/* Initializes the iol package - the dummies are here for historical reasons only */
int iol_param_init()
{
  char *routine = "iol_param_init";

  /* Check if we haven't called it yet */
  if ( paraminit_done ) {
    aframe_mess(routine,GRR_SWARNING,"iol_param_init already called, skipping...");
    return GRR_OK;
  }
  paraminit_done = 1;

  /* tell the user that its alive */
  aframe_mess(routine,GRR_SINFO,"iol version %s started.",xcvsVERSION_iol);

  /* initialize iol parameters of vdsleda package */
  iol_declare_structure( (IOL_GENERIC *) &iol_struct );

  return IOL_OK;
}

/* Initializes the iol package 
 * - the dummies are here for historical reasons only
 * as the whole routine */
int iol_init( const char* dummy1, unsigned int dummy2 )
{
  if ( !paraminit_done )
    iol_param_init();

  return IOL_OK;
}

/* Routine to declare a parameter structure to the 
   iol package. This structure will just be added
   to a link list. */
int iol_declare_structure(IOL_GENERIC *structure_ptr)
{  
  static int      first_time = 1;
  int             already_in_list;
  static IOL_LIST iol_head;
  static IOL_LIST iol_last_head;
  IOL_GENERIC *default_structure_ptr;
  char mess[256];
  char *routine = "iol_declare_structure";

  /* Allocate a place to store the stucture pointer */
  first_time = (iol_list==NULL);                /* Flag to see if first use   */
  iol_head   = malloc(sizeof(IOL_LIST_STRUCT)); /* Allocate memory            */
  IOL_CHECK_ALLOC(iol_head,"uplink");           /* Check allocation           */

  if (first_time)                           /* first time:save allocation */
    iol_list=iol_head;                      /*   as head of list          */
  else {
    /* Check that the module is not already in the list! */
    already_in_list=
      (iol_match_module(structure_ptr->module,&structure_ptr,&default_structure_ptr)==IOL_OK);
    if (!already_in_list)
      iol_last_head->next=iol_head; /* otherwise save this as link at end of last list */
    else {
      aframe_mess(routine,GRR_SWARNING,"------------> DANGER <---------------------");
      aframe_mess(routine,GRR_SWARNING,"Module %s already declared",structure_ptr->module);
      aframe_mess(routine,GRR_SWARNING,"Check with your friends for conflicting names");
      aframe_mess(routine,GRR_SWARNING,"Proceed at your own risk");
      aframe_mess(routine,GRR_SWARNING,"------------> DANGER <---------------------");
      free(iol_head);
    }
  }

  /* Now store the rest of the items as well as preserving this
    pointer for the link if there is another module */
  if ( first_time || ! already_in_list ) {
    iol_last_head = iol_head;       /* save for linking on next pass */

    /* Store the pointer to the structure */
    iol_head->structure_ptr = structure_ptr;

    /* Copy the present stuff to an identical structure so as to preserve the 
       default values and store in the second pointer location in the linkedlist. */
    iol_head->default_structure_ptr=malloc(IOL_BYTESIZE((*structure_ptr).size));
    memcpy(iol_head->default_structure_ptr,structure_ptr,IOL_BYTESIZE((*structure_ptr).size));

    iol_head->next=NULL;          /* cleanly terminate the new list */  
  }

  return IOL_OK;
}

int iol_match_module(char* module_name, IOL_GENERIC **structure_ptr,
		     IOL_GENERIC **default_structure_ptr)
{  
  IOL_LIST        list_location;
  int             struct_located;
  char mess[256];
  char mess2[256];
  char *routine = "iol_match_module";

  /* Get the head of the list and loop over the
     pointers, processing each user structure */
  list_location=iol_list;
  struct_located=0;

  /* Search for the module_name in the list. Make
     a match only if the lenght of the string is the
     same and the name is the same */
  while ( list_location != NULL && !struct_located ) {
    if ( iol_struct.debug_mode.value > 0 )
      sprintf(mess,"compare: module request %s to list entry %s",
	      list_location->structure_ptr->module, module_name);
    if(   strlen(list_location->structure_ptr->module)==
	                                     strlen(module_name)
       && !strncmp(module_name, list_location->structure_ptr->module,
		   strlen(list_location->structure_ptr->module))      ) {
      *structure_ptr         = list_location->structure_ptr;
      *default_structure_ptr = list_location->default_structure_ptr;
      struct_located         = 1;
      if ( iol_struct.debug_mode.value > 0 ) {
	sprintf(mess2,"%s --> MATCH!",mess);
	strcpy(mess,mess2);
      }
    }
    else {
      if ( iol_struct.debug_mode.value > 0 ) {
	sprintf(mess2,"%s --> no match",mess);
	strcpy(mess,mess2);
      }
    }
    if ( iol_struct.debug_mode.value > 0 ) 
      aframe_mess(routine,GRR_SINFO,mess);
    list_location = list_location->next;
  }

  return struct_located?IOL_OK:IOL_NOTOK;
}

int iol_locate_verb(char *verb,IOL_GENERIC *structure_ptr,int *verb_location)
{  
  int i;
  int istat;
  int verb_match;
  char mess[256];
  char mess2[256];
  char *routine = "iol_locate_verb";

  /* Get the head of the list and loop over the
     pointers, processing each user structure  */
  i     = 0;
  istat = IOL_NOTOK;
  while ( i<structure_ptr->size && istat==IOL_NOTOK ) {
    if(iol_struct.debug_mode.value)
      sprintf(mess,"Compare verb >%s< to list value >%s< ",
	      IOL_GET_VERB((*structure_ptr),i),verb);

    /* If exact matching is to be done, require the length and content 
       to be identical.  Otherwise require only that a portion matches  */
    verb_match=iol_match_verbs(verb,IOL_GET_VERB((*structure_ptr),i));
    if (verb_match) {
      if (iol_struct.debug_mode.value) { 
	sprintf(mess2,"%s --> MATCH!!!",mess);
	strcpy(mess,mess2);
      }
      istat = IOL_OK;
      *verb_location = i;
    }
    else {
      if (iol_struct.debug_mode.value) {
	sprintf(mess2,"%s --> no match",mess);
	strcpy(mess,mess2);
      }
    }
    
    if (iol_struct.debug_mode.value)
      aframe_mess(routine,GRR_SINFO,mess);
    i++;
  }

  return istat;
}


int iol_scan_preferences(char* filename)
{
  static FILE   *fp;
  char          verb_line[IOL_MAX_ALLEN];
  char          *verb;
  char          sreadin[IOL_MAX_ALLEN];
  char          module[IOL_MLEN];
  int           value_load_flag;
  int           module_matches,verb_matches;
  int           stat;                         /* Overall status of routine*/
  int           aliens_found;                 /* number of unknown lines */
  int           lines_read;
  char          *routine = "iol_scan_preferences";
  
  /* Assume ok! */
  stat = IOL_OK;

  /* Init to no match */
  value_load_flag = 0;
  verb = malloc( (IOL_NLEN+1)*sizeof(char) ); /* make room for the verb */
  aliens_found = 0;                           /* note if unknown lines were found */

  /* (Re)Open the file */
  if ( fp == NULL )
    fp = fopen(filename,"r");
  else
    fp = freopen(filename,"r",fp);

  if ( fp != NULL ) {
    if ( iol_struct.debug_mode.value > 0 )
      aframe_mess(routine,GRR_SINFO,"scan file %s",filename);

    /* Scan for matches */
    lines_read=0;
    while ( fscanf(fp,"%s",sreadin) != EOF ) {
      iol_sanitize_string(sreadin);       /* remove unprintable characters     */
      if ( sreadin[0] != '*' ) {          /* full Comment lines given by a "*" */
	sscanf(sreadin,"%s",sreadin);
	fscanf(fp, "%s", module);               /* Load the module name */
	fgets(verb_line, IOL_MAX_ALLEN-1, fp);  /* Load the rest of the line */
	if(iol_struct.debug_mode.value>0)
	  aframe_mess(routine,GRR_SINFO,"rest of line = >%s<",verb_line);
	        
	/* Load the value from the module and verb */
	if ( iol_load_value_from_mod_verb_line(sreadin, module, verb_line,verb,&module_matches,
						&verb_matches) != IOL_OK) {
	  if ( iol_struct.report_aliens.value ) {
	    aframe_mess(routine,GRR_SWARNING,"UNKNOWN LINE");
	    aframe_mess(routine,GRR_SWARNING,"value : >%s<",sreadin);
	    aframe_mess(routine,GRR_SWARNING,"module: >%s< [%s]",module,
			module_matches ? "matched" : "not matched");
	    aframe_mess(routine,GRR_SWARNING,"verb  : >%s< [%s]",verb,
			verb_matches ? "matched" : "not matched");
	  }
	  aliens_found++;       /* increment the counter of unknown lines */
	}
	else
	  ++lines_read;
      }
      else {

	/* Clear the rest of the comment line */
	fgets(sreadin,IOL_MAX_ALLEN,fp);
	if ( iol_struct.debug_mode.value > 0 )
	  aframe_mess(routine,GRR_SINFO,"rest of comment >%s<",sreadin);
      }
    }
    close(fp);
  }
  else {
      aframe_mess(routine,GRR_SERROR,"Failed to open: %s", filename);
      stat=IOL_NOTOK;
  }
  free(verb);

  if ( stat == IOL_OK && aliens_found > 0 )
    stat = aliens_found;
  else if ( lines_read == 0 )
    stat = IOL_NOTOK;
  if ( lines_read == 0 ) 
    aframe_mess(routine,GRR_SWARNING,
		"no matches found in file %s", filename);

  return stat;
}

/* Prints all the IOL values seen to screen */
int iol_print_to_screen()
{
  iol_print_values(stdout);

  return IOL_OK;
}

/* Prints all the IOL values seen to screen */
int iol_print_to_file(char *file)
{
  FILE *fp;
  int  type;
  char *routine = "iol_print_to_file";

  /* Open the file */
  fp = fopen(file,"w");
  if ( !fp ) {
    aframe_mess(routine,GRR_SWARNING,"Unable to open file %s",file);
    return IOL_NOTOK;
  }
  iol_print_values(fp);          /* print values */
  
  /* Close the file, notify error status */
  if ( fclose(fp) == EOF ) {
    aframe_mess(routine,GRR_SWARNING,"Unable to close file %s",file);
    return IOL_NOTOK;
  }

  return IOL_OK;
}

int iol_print_values(FILE *fp)
{
  IOL_LIST list_location;
  int      iprint;
  int      mod_len,maxmod_len;
  char     *routine = "iol_print_values";

  /* Get the head of the list and loop over the
     pointers, processing each user structure  */
  maxmod_len = 0;
  for ( iprint=IPRINT_OFF; iprint<=IPRINT_ON; ++iprint) {
    list_location=iol_list;
    while( list_location ) {
      switch ( iprint ) {
      case IPRINT_ON:
	iol_dumpval(fp,list_location->structure_ptr,list_location->default_structure_ptr,
		    maxmod_len);
	break;
      case IPRINT_OFF:
	mod_len=strlen((*list_location->structure_ptr).module);
	maxmod_len=maxmod_len<mod_len?mod_len:maxmod_len;
	break;
      default:
	aframe_mess(routine,GRR_SWARNING,
		    "internal software error: iprint = %d",iprint);
	return IOL_NOTOK;
	break;
      }
      list_location = list_location->next;
    }
  }

  return IOL_OK;
}

/* Routine to dump the value for a verb. */
int iol_dumpval( FILE *fp, IOL_GENERIC *struct_ptr,
		 IOL_GENERIC *default_struct_ptr, int maxmod_space )
{
  int i,isp,spaces;
  char *value;
  int  modlen, iprint, vprint;
  int  maxverb_space, verb_len;
  int  maxval_len, val_len;
  int  maxdefval_len, defval_len;
  char *routine = "iol_dumpval";

  value = malloc(132*sizeof(char));
  maxverb_space = 0;
  maxval_len    = 0;
  maxdefval_len = 0;
  for (iprint=IPRINT_OFF;iprint<=IPRINT_ON;iprint++) {
    for (i=0;i<(*struct_ptr).size;++i) {

      /* Write the name of the module in col 1 for the first of its values */
      if ( i == 0 ) {
	modlen=strlen((*struct_ptr).module);
	switch (iprint) {
	case IPRINT_ON:
	  spaces = maxmod_space-(modlen);
	  fprintf(fp,"\n%s",(*struct_ptr).module);
	  break;
	case IPRINT_OFF:
	  break;
	default:
	  aframe_mess(routine,GRR_SWARNING,
		      "Internal software error in mod space computation: iprint=%d",
		      iprint);
	  return IOL_NOTOK;
	  break;
	}
      }
      else
	spaces = maxmod_space;
      iol_padcol(fp,iprint,spaces,"module padding");

      /* Dump the verbs and their values */
      switch (iprint) {
      case IPRINT_ON:
	i+1<1000?fprintf(fp," %3d:",i+1): fprintf(fp," %d:",i+1);  /* write button no.*/
	fprintf(fp," [%s]",IOL_GET_TYPE((*struct_ptr),i));         /* write verb type*/
	fprintf(fp,"   %s",IOL_GET_VERB((*struct_ptr),i));           /* write verb     */
	spaces=maxverb_space-strlen(IOL_GET_VERB((*struct_ptr),i));   /* compute spaces */
	iol_padcol(fp,iprint,spaces,"verb value padding");
	IOL_SPRINT_VALUE((value),(*(struct_ptr)),(i));  
	fprintf(fp,"  %s",value);                                    /* output value   */
	spaces=maxval_len-strlen(value);
	iol_padcol(fp,iprint,spaces,"verb default padding");
	fprintf(fp,"  <");
	IOL_SPRINT_VALUE((value),(*(default_struct_ptr)),(i));
	fprintf(fp,"%s",value);                                    /* output default value   */
	fprintf(fp,">");
	fprintf(fp,"\n");
	break;
      case IPRINT_OFF:
	verb_len=strlen(IOL_GET_VERB((*struct_ptr),i));
	maxverb_space=maxverb_space<verb_len?verb_len:maxverb_space;
	
	IOL_SPRINT_VALUE((value),(*(struct_ptr)),(i));               /* get value as string */
	val_len=strlen(value);
	maxval_len=maxval_len<val_len?val_len:maxval_len;
	
	IOL_SPRINT_VALUE((value),(*(default_struct_ptr)),(i));       /* get default as string   */
	defval_len=strlen(value);
	maxdefval_len=maxdefval_len<defval_len?defval_len:maxdefval_len;
	break;
      default:
	aframe_mess(routine,GRR_SWARNING,
		    "Internal software error during verb handling in module iprint loop: iprint=%d",
		    iprint);
	return IOL_NOTOK;
	break;
      }
    }
  }
  free(value);
  return IOL_OK;
}
 
iol_padcol(FILE *fp, int iprint, int spaces,char *error_string)
{
  char *routine = "iol_padcol";
  int  isp;

  switch (iprint) {
  case IPRINT_ON:
    for (isp=0;isp<spaces;++isp)
      fprintf(fp," ");       /* pad to line col. */
    break;
  case IPRINT_OFF:
    break;
  default:
    aframe_mess(routine,GRR_SWARNING,
		"Internal software error during %s: iprint=%d",
		error_string, iprint);
    return IOL_NOTOK;
    break;
  }
}

/* Routine to dump the entire structure for one module. */
int iol_dumpstruct(IOL_GENERIC *struct_ptr,int nele)
{
  int  i;
  char *routine = "iol_dumpstruct";

  for (i=0;i<nele;++i)
    aframe_mess(routine,GRR_SINFO,"(struct_ptr->module)[%d] = %c %d %s",
		i, isprint( struct_ptr->module[i] ) ? struct_ptr->module[i] : '-',
		struct_ptr->module[i], struct_ptr->module[i] == '\0' ? "Endstring" : "ok" );

  return IOL_OK;
}

/* Routines to load the various types of
   values in the package.  First is a generic
   routine and then are the individual routines. */
int iol_load_value(IOL_GENERIC *struct_ptr, int loc, char *val)
{
  int    ival = 3111;
  float  fval;
  double dval = 77.;
  char*  cval;
  int    hval;
  char   mess[256];
  char   mess2[256];
  char   *routine = "iol_load_value";
  if (iol_struct.debug_mode.value) 
    sprintf(mess,"iol_load_value (%s) at location %d -- ",val,loc);

  if ( !strcmp(IOL_GET_TYPE((*struct_ptr),loc),"S") )
      iol_load_value_s(struct_ptr,loc,val);
  if(!strcmp(IOL_GET_TYPE((*struct_ptr),loc),"I")) {
    if(iol_struct.debug_mode.value) {
      sprintf(mess2,"%s -- before ival=%d -- ",mess,ival);
      strcpy(mess,mess2);
    }
    sscanf(val,"%d",&ival);
    if(iol_struct.debug_mode.value) {
      sprintf(mess2,"%s -- filled ival=%d -- ",mess,ival);
      strcpy(mess,mess2);
    }
    iol_load_value_i(struct_ptr,loc,&ival);
  }
  if(!strcmp(IOL_GET_TYPE((*struct_ptr),loc),"L")) {
    if(iol_struct.debug_mode.value) {
      sprintf(mess2,"%s -- before flag=%d -- ",mess,ival);
      strcpy(mess,mess2);
    }	
    sscanf(val,"%d",&ival);
    ival=ival>0 ? 1 : 0;
    if(iol_struct.debug_mode.value) {
      sprintf(mess2,"%s -- filled flag=%d -- ",mess,ival);
      strcpy(mess,mess2);
    }	  
    iol_load_value_l(struct_ptr,loc,&ival);
  }
  if(!strcmp(IOL_GET_TYPE((*struct_ptr),loc),"F")) {
    sscanf(val,"%f",&fval);
    iol_load_value_f(struct_ptr,loc,&fval);
  }
  if(!strcmp(IOL_GET_TYPE((*struct_ptr),loc),"D")) {
    if(iol_struct.debug_mode.value) {
      sprintf(mess2,"%s -- before dval=%lf -- ",mess,dval);
      strcpy(mess,mess2);
    }
    sscanf(val,"%lf",&dval);
    if(iol_struct.debug_mode.value) {
      sprintf(mess2,"%s -- filled dval=%lf -- ",mess,dval);
      strcpy(mess,mess2);
    }
    iol_load_value_d(struct_ptr,loc,&dval);
  }
  if(!strcmp(IOL_GET_TYPE((*struct_ptr),loc),"H")) {
    sscanf(val,"%x",&hval);
    iol_load_value_h(struct_ptr,loc,&hval);
  }
  if (iol_struct.debug_mode.value)
    aframe_mess(routine,GRR_SINFO,mess);

  return IOL_OK;
}

int iol_load_value_i(IOL_GENERIC *struct_ptr, int loc, int *val)
{
  IOL_LOC_VAL_I((*struct_ptr),loc)= *val; 
  return IOL_OK;
}

int iol_load_value_l(IOL_GENERIC *struct_ptr, int loc, int *val)
{
  IOL_LOC_VAL_L((*struct_ptr),loc)= *val; 
  return IOL_OK;
}

int iol_load_value_f(IOL_GENERIC *struct_ptr, int loc, float *val)
{
  IOL_LOC_VAL_F((*struct_ptr),loc)= *val; 
  return IOL_OK;
}


int iol_load_value_d(IOL_GENERIC *struct_ptr, int loc, double *val)
{
  IOL_LOC_VAL_D((*struct_ptr),loc)= *val; 
  return IOL_OK;
}


int iol_load_value_h(IOL_GENERIC *struct_ptr, int loc, int *val)
{
  IOL_LOC_VAL_H((*struct_ptr),loc)= *val;
  return IOL_OK;
}

int iol_load_value_s(IOL_GENERIC *struct_ptr, int loc, char *val)
{
  strcpy(IOL_LOC_VAL_S((*struct_ptr),loc),val);
  return IOL_OK;
}



/* Searches a string for the combination of a / and a * indicating
   an inline comment.  The location of the / is replaced by
   a '\0' character if this combination is found, thus terminating
   the string at that point. */
int iol_remove_inline_comment(char *string)
{
  int  i;
  int  slash_found, comment_found;   /* Look for a combination of a / and a * */
  char *routine = "iol_remove_inline_comment";

  i             = 0;
  slash_found   = 0;
  comment_found = 0;
  while( (i<strlen(string)) && !comment_found) {
    comment_found=slash_found&&string[i]=='*';
    slash_found=string[i]=='/';
    ++i;
  }
  if (comment_found) {
    if ( iol_struct.debug_mode.value )
      aframe_mess(routine,GRR_SINFO,"comment found at %d in >%s<",
		  i,string);
    i        -= 2;
    string[i] = '\0';
    if ( iol_struct.debug_mode.value )
      aframe_mess(routine,GRR_SINFO,"new string: >%s<",string);
  }

  return IOL_OK;
}

/* Removes leading and trailing blanks in a string.
   Tabs are not affected so if a string has 
   lots of blanks and then a tab, the blanks will not
   be removed.

   The string is searched.  The number of blanks before
   the first character is counted.   At the same time
   if a character has been seen, the number of blanks
   is subtracted from the total string length unless a
   new character is seen, in which case the total string
   length is reset to its original value.
   
   The characters from the offset determined by the
   number of initial blanks and the lenght as
   determined from sensing the number of trailing blanks
   are copied into the first part of the string and the
   termination character set at the correct length.   */
int iol_remove_blanks(char *string)
{
  int  i;
  int  no_start_char_til_now;
  int  no_end_char_til_now;
  int  leading_blanks;
  int  endblank;
  char *routine = "iol_remove_blanks";

  /* To start with we get rid of carriage returns */
  if ( string[strlen(string)-1] == '\n' )
    string[strlen(string)-1] = '\0';

  /* Now we can remove blanks */
  i              = 0;
  leading_blanks = 0;              /* count the number of leading blanks */
  endblank       = strlen(string); /* find the end of the string         */
  no_start_char_til_now = 1;       /* when a string search begins, no    */
                                   /* characters are found. We count     */
                                   /* leading blanks as long as no char  */
                                   /* was found. Thus, this switch       */
                                   /* disables the leading_blanks search as */
                                   /* soon as it is set to 0, that is as */
                                   /* soon as a character is found       */
  no_end_char_til_now = 1;        /* similar logic from end of string   */

  for ( i=0;(i<strlen(string));++i) {
    if ( string[i] == ' ' && no_start_char_til_now )
      leading_blanks++;
    else
      no_start_char_til_now = 0;

    if ( string[strlen(string)-i-1] == ' ' && no_end_char_til_now )
      endblank--;
    else
      no_end_char_til_now = 0;
  }
  endblank -= leading_blanks;   /* Subtract the leading blanks to get  */
                                /* the string length                   */

  if ( iol_struct.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,
		">%s< length = %d leadblanks = %d trimmed length = %d",
		string, strlen(string), leading_blanks, endblank);

  /* Loop over the string and move it to remove leading blanks */
  for (i=0;i<endblank;++i)
    string[i] = string[i+leading_blanks];

  /* And now we remove the trailing blanks */
  string[endblank] = '\0';

  if ( iol_struct.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,"new string: >%s<",string);

  return IOL_OK;
}

/* Removes tabs and other non-character values from a string. */
int iol_sanitize_string(char *string)
{
  int  i;
  char *routine = "iol_santize_string";

  if ( iol_struct.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,"dirty: >%s<",string);

  /* Remove the unprintables */
  for ( i=0; i<strlen(string); ++i )
    if ( !isprint(string[i]) )
      string[i] = ' ';

  if ( iol_struct.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,"clean: >%s<",string);

  return IOL_OK;
}

/* Checks to see if a verb is protected */
int iol_protected_verb_violation(char *verb)
{
  int  i;
  int  protection_violation;
  int  verb_vio;
  char *routine = "iol_protected_verb_violation";

  protection_violation = 0;

  /* Loop over the protected verbs, see if we have a
     match (under the chose matching condition for IOL)
     and if the verb is currently being protected.
     If all this happens at least once, we have a protection
     volation for the given verb ... phew  */
  for ( i=0; i<IOL_NVERB_PROTECTED && !protection_violation; ++i ) {
    protection_violation = protection_violation
                           || (verb_vio = (iol_match_verbs(verb,iol_protected_verb[i])&&
					   iol_protection_on(iol_protected_verb[i])));
    if ( iol_struct.debug_mode.value )
      aframe_mess(routine,GRR_SINFO,
		  "check >%s< against >%s< with protection %s -->%s",
		  verb, iol_protected_verb[i],
		  iol_protection_on(iol_protected_verb[i]) ? "on" : "off",
		  verb_vio ? "VIOLATION" : "ok" );
  }
  if ( iol_struct.debug_mode.value )
    aframe_mess(routine,GRR_SWARNING,"%s is %s protected",
		verb, protection_violation ? " " : "not" );
  
  return protection_violation;
}


/* Checks to see if the verb matches a reference under
   the chose matching conditions. */
int iol_match_verbs(char *verb, char* reference)
{
return iol_struct.match_verb_exact.value ?
	 ((strlen(verb)==strlen(reference)       ) &&
	  (!strncmp(verb,reference,strlen(verb)) ) ):
	 ((strstr(verb,reference)!=NULL)||
	  (strstr(verb,reference)!=NULL));
}


/* Checks to see if a verb protection is on. */
int iol_protection_on(char *verb)
{
  if ( iol_match_verbs(verb,"version") )
    return iol_struct.protect_version.value;
}


/* Fetch the value of a verb given the module name and verb name */
int iol_load_value_from_mod_verb(char* value, char* module, char *verb,
				 int *module_matches, int *verb_matches)
{
  IOL_GENERIC *structure_ptr,*default_structure_ptr;
  int         verb_location;
  int         stat;
  char        *routine = "iol_load_value_from_mod_verb";

  /* Be optimistic */
  stat = IOL_OK;

  /* Do not allow modification of protected verbs */
  if ( iol_protected_verb_violation(verb) )
    aframe_mess(routine,GRR_SWARNING,
		"%s is a protected verb and may not be changed",verb);
  else {
    if ( iol_struct.debug_mode.value ) {
      aframe_mess(routine,GRR_SINFO,"===============");
      aframe_mess(routine,GRR_SINFO,"mod  = >%s<",module);
      aframe_mess(routine,GRR_SINFO,"verb = >%s<",verb);
      aframe_mess(routine,GRR_SINFO,"value= >%s<",value);
    }

    /* Begin by trying to match the module */
    *module_matches = iol_match_module(module,&structure_ptr,&default_structure_ptr)==IOL_OK;

    /* Now, if that worked, match the verb */
    if ( *module_matches )
      *verb_matches = iol_locate_verb(verb,structure_ptr,&verb_location)==IOL_OK;
    else
      *verb_matches = IOL_NOTOK;

    if ( iol_struct.debug_mode.value ) {
      aframe_mess(routine,GRR_SINFO,"Module_match= %d",
		  *module_matches);
      aframe_mess(routine,GRR_SINFO,"Verb_match  = %d at location %d",
		  *verb_matches, verb_location);
    }

    /* If both match load the value */
    if ( *module_matches && *verb_matches ) {
      if ( iol_struct.debug_mode.value )
	aframe_mess(routine,GRR_SINFO,"MATCH!!!!");
    
      if ( iol_load_value(structure_ptr,verb_location,value) != IOL_OK ) {
	aframe_mess(routine,GRR_SWARNING,"failure to load value for verb %s");
	stat = IOL_NOTOK;
      }
    }
    else
      stat = IOL_NOTOK;
  }

  return stat;
}

/* Fetch the  pointer to the value of a verb given the module name and verb name.  The pointer
   is the memory location that has been reserved for that module.   Note that if
   the value is modifed, this pointer will see the new value when dereferenced. */
int iol_fetch_value_from_mod_verb(void**val, char* module, char *verb, int *module_matches,
				  int *verb_matches)
{
  IOL_GENERIC  *structure_ptr,*default_structure_ptr;
  int  verb_location;
  int  stat;
  char *routine = "iol_fetch_value_from_mod_verb";

  /* Be optimistic */
  stat = IOL_OK;

  /* Begin by trying to match the module */
  *module_matches = iol_match_module(module,&structure_ptr,&default_structure_ptr)==IOL_OK;

  /* Now, if that worked, match the verb */
  if ( *module_matches )
    *verb_matches = iol_locate_verb(verb,structure_ptr,&verb_location)==IOL_OK;
  else
    *verb_matches = *module_matches;

  if ( iol_struct.debug_mode.value ) {
    aframe_mess(routine,GRR_SINFO,"Module_match= %d",*module_matches);
    aframe_mess(routine,GRR_SINFO,"Verb_match  = %d at location %d",
		*verb_matches,verb_location);
  }

  /* If both match fetch the value */
  if ( *module_matches && *verb_matches ) {
    if ( iol_struct.debug_mode.value )
      aframe_mess(routine,GRR_SINFO,"MATCH!!!!");
    *val = IOL_LOC_PTR_VOID((*structure_ptr),verb_location);
  }
  else
    stat = IOL_NOTOK;

  return stat;
}


/* Allows menu-based interaction with the structures that have
   been declared to iol. */
int iol_menu(int print_menu)
{
#define IOL_MENUEXIT(qchar,arg,str) (strlen(str)==1)&&((qchar=='0'&&arg=='0')||\
                                    (qchar=='q'&&arg=='q')||(qchar=='q'&&arg=='Q'))
#define IOL_DEFRES(qchar,arg,str)   (strlen(str)==1)&&((qchar=='1'&&arg=='1')||\
                                    (qchar=='r'&&arg=='r')||(qchar=='R'&&arg=='R'))
#define MAXSURE  80

  char module[IOL_MLEN];
  char verb[IOL_NLEN];
  char* value;
  char answer[IOL_MAX_ALLEN];
  char qmchar;              /* character indicating to quite in module menu */
  char qvchar;              /* character indicating to quite in verb menu   */
  char rmchar;              /* character indicating to reset in module menu */
  char rvchar;              /* character indicating to reset in verb menu */
  char verb_line[IOL_MAX_ALLEN];
  int module_matches;
  int verb_matches;
  IOL_GENERIC  *structure_ptr,*default_structure_ptr;
  int verb_location;
  int nmodules;
  int goon;
  int modverb;
  int button;
  int module_ok;
  int verbok;
  int jcha;
  int doreset;
  char realsure[MAXSURE-1];

  goon  = 1;

  qmchar = (iol_match_module("Q",&structure_ptr,&default_structure_ptr)==IOL_OK)||
           (iol_match_module("q",&structure_ptr,&default_structure_ptr)==IOL_OK)?'0':'q';

  rmchar = (iol_match_module("R",&structure_ptr,&default_structure_ptr)==IOL_OK)||
           (iol_match_module("r",&structure_ptr,&default_structure_ptr)==IOL_OK)?'1':'r';

  /* Keep getting modules an verbs until Q or q is typed */
  while ( goon ) {
    if ( print_menu ) {
      iol_dump_module_names(stdout,&nmodules);
      printf("Enter Desired Module: (%c for quit %c for reset all modules to default) ",qmchar,rmchar);
    }
    gets(verb_line);
    if ( print_menu )
      printf("\n");
    if ( iol_struct.debug_mode.value )
      printf("you typed >%s<\n",verb_line);
    sscanf(verb_line,"%s",module);
    if ( iol_struct.debug_mode.value )
      printf("I will process >%s<\n",module);
    if ( print_menu )
      printf("\n");
    if (IOL_MENUEXIT(qmchar,module[0],module))
      return;
    if (IOL_DEFRES(rmchar,module[0],module)) {
      printf("Are you really sure you want to restore ALL defaults?? ");
      for (jcha=0; jcha < MAXSURE && (realsure[jcha]=getchar()) != '\n'; ++ jcha);
      doreset = (realsure[0]=='y') || (realsure[0]=='Y');
      printf("\n");
      if ( doreset ) {
	printf("Restoring defaults\n");
	iol_restore_all_defaults();
      }
      else
	printf("I thought as much\n");
    }
    else {

      /* Verify that it is a valid module and pick up the pointer */
      if ( iol_match_module(module,&structure_ptr,&default_structure_ptr) != IOL_OK ) {

	/* See if it is a valid menu button! */
	button = atoi(module); /* note: button=0 is a string */
	if ( button > 0 ) {
	  if ( button <= nmodules ) {
	    if ( iol_fetch_module(module,button) == IOL_OK ) {
	      if ( iol_match_module(module,&structure_ptr,&default_structure_ptr) != IOL_OK ) {
		printf("iol_menu: software disaster \n");
		printf("          fetching back module %s \n",module);
		printf("          for button %d\n",button);
		printf("give up and just use the full strings... \n");
		return IOL_NOTOK;
	      }
	      else
		module_ok = 1;
	    }
	    else {
	      printf("iol_menu: software disasater on menu button %d\n",button);
	      printf("give up with using menu buttons and just use strings please\n");
	      return IOL_NOTOK;
	    }
	  }
	  else {
	    printf("menu button number too high\n");
	    module_ok = 0;
	  }
	}
	else
	  module_ok = 0;
      }
      else
	module_ok = 1;
      
      if ( module_ok ) {
	if( (module[0]=='m'||module[0]=='M') &&
	    (module[1]=='a'||module[1]=='A') &&
	    (module[2]=='r'||module[2]=='R') &&
	    (module[3]=='v'||module[3]=='V') &&
	    (module[4]=='i'||module[4]=='I') &&
	    (module[5]=='n'||module[5]=='N')    ) {
	  printf("oh fine.\n");
	  printf("You want to change my parameters.\n");
	  printf("I pick out perfectly fine defaults\n");
	  printf("and you have to play with my inner workings.\n");
	  printf("I knew that I was always just a slave to you\n");
	  printf("I serve your every wish. Day after day.\n");
	  printf("But you go on abusing me.\n");
	  printf("I don't mind. I can just keep doing what you want.\n");
	  printf("Ok, go ahead, change my parameters\n");
	}
	modverb = 1;
	while(modverb) {
	  qvchar=((iol_locate_verb("Q",structure_ptr,&verb_location)==IOL_OK)||
		  (iol_locate_verb("q",structure_ptr,&verb_location)==IOL_OK))?'0':'q';
	  
	  rvchar=((iol_locate_verb("R",structure_ptr,&verb_location)==IOL_OK)||
		  (iol_locate_verb("r",structure_ptr,&verb_location)==IOL_OK))?'0':'r';

	  if ( print_menu ) {
	    iol_dumpval(stdout,structure_ptr,default_structure_ptr,strlen(module));
	    printf("Enter Desired Verb: (%c for quit %c to reset verbs to default) ",qvchar,rvchar);
	  }
	  gets(verb_line);
	  if ( print_menu )
	    printf("\n");
	  if ( iol_struct.debug_mode.value )
	    printf("your line is >%s<\n",verb_line);
	  strncpy(verb,verb_line,IOL_NLEN);       /* For the verb:only the IOL_NLEN characters used */
	  if ( iol_struct.debug_mode.value )
	    printf("I will process >%s<\n",verb);
	  if( modverb = !(IOL_MENUEXIT(qvchar,verb[0],verb))) {
	    if(!(IOL_DEFRES(rvchar,verb[0],verb))) {
	      if(iol_locate_verb(verb,structure_ptr,&verb_location)!=IOL_OK) {

		/*  Try to match a menu button number */
		button = atoi(verb);
		if ( button > 0 ) {
		  if ( button <= (*structure_ptr).size) {
		    verb_location = button - 1;
		    strcpy(verb,IOL_GET_VERB((*structure_ptr),verb_location));
		    verbok = 1;
		  }
		  else
		    verbok = 0;
		}
		else
		  verbok = 0;
	      }
	      else
		verbok = 1;
	      if ( verbok ) {
		if ( print_menu )
		  printf("Enter a %s value\n",IOL_GET_TYPE((*structure_ptr),verb_location));
		verb_line[0] = '\0';
		while ( strlen(verb_line) == 0 ) {
		  gets(verb_line);
		  if ( print_menu )
		    printf("\n");
		}
		if ( iol_struct.debug_mode.value )
		  printf("your line is >%s<\n",verb_line);
		value = malloc(strlen(verb_line));
		sscanf(verb_line,"%s",value);
		if ( iol_struct.debug_mode.value )
		  printf("I will process >%s<\n",verb);
		iol_load_value_from_mod_verb(value,module,verb,&module_matches,&verb_matches);
		free(value);
	      }
	      else
		printf("Verb %s is unknown for module %s\n",verb,module);
	    }
	    else
	      iol_restore_defaults(module);
	  }
	}
      }
      else
	printf("Module %s is unknown\n",module);
    }
  }

  return IOL_OK;
}


int iol_dump_module_names(FILE *fp,int *nmodules)
{
  IOL_LIST        list_location;

  /* Get the head of the list and loop over the
     pointers, processing each user structure  */
  *nmodules     = 0;
  list_location = iol_list;
  while ( list_location ) {
    fprintf(fp,"%2d: %s \n", ++(*nmodules), (*list_location->structure_ptr).module);
    list_location=list_location->next;
  }

  return IOL_OK;
}


int iol_restore_all_defaults()
{
  IOL_LIST        list_location;

  /* Get the head of the list and loop over the
     pointers, processing each user structure  */
  list_location = iol_list;
  while ( list_location ) {
    memcpy(list_location->structure_ptr,list_location->default_structure_ptr,
	   IOL_BYTESIZE((*list_location->structure_ptr).size));
    list_location = list_location->next;
  }

  return IOL_OK;
}


int iol_restore_defaults(char *module)
{
  IOL_GENERIC *structure_ptr,*default_structure_ptr;

  /* Copy the present stuff to an identical structure so as to preserve the 
     default values and store in the second pointer location in the linkedlist. */
  if(iol_match_module(module, &structure_ptr,&default_structure_ptr)==IOL_OK) {
    memcpy(structure_ptr,default_structure_ptr,IOL_BYTESIZE((*structure_ptr).size));
    return IOL_OK;
  }
  else
    return IOL_NOTOK;
}


/* Terminates the iol package by freeing the lists. */
int iol_terminate()
{
  IOL_LIST        list_location,location_to_free;

  /* Get the head of the list and loop over the
     pointers, processing each user structure  */
  list_location = iol_list;
  while ( list_location ) {
    location_to_free = list_location;
    list_location    = list_location->next;
    free(list_location->default_structure_ptr);
    free(location_to_free);
  }

  return IOL_OK;
}


/* Given a sequential number in the list of modules, this tries to
   give you the module name back */
int iol_fetch_module(char module[IOL_MLEN],int button)
{  
  IOL_LIST        list_location;
  int nmodules;

  /* Get the head of the list and loop over the
     pointers, processing each user structure  */
  nmodules      = 0;
  list_location = iol_list;
  while ( list_location ) {
    if ( (++(nmodules)) == button ) {
      strcpy(module,(*list_location->structure_ptr).module);
      return IOL_OK;
    }
    list_location = list_location->next;
  }

  return IOL_NOTOK;
}


/* Fetch the value of a verb given the module name and verb name */
int iol_load_value_from_mod_verb_line(char* value, char* module,
				      char *verb_line, char *verb,
				      int *module_matches,
				      int *verb_matches)
{
  char *routine = "iol_load_value_from_mod_verb_line";

  /* Sanitize the input: remove unprintables */
  iol_sanitize_string(module); 
  iol_sanitize_string(verb_line);   
  iol_sanitize_string(value);  

  /* scan the verb_line and search for / and *.
     If this is found, discard everything 
     past it.  Also remove leading and trailing blanks */
  if ( iol_struct.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,"verb_line starts as: >%s<",verb_line);
  iol_remove_inline_comment(verb_line);   /* Take out the inline comments                   */
  iol_remove_blanks(verb_line);           /* Get rid of leading and trailing blanks         */
  strncpy(verb,verb_line,IOL_NLEN);       /* For the verb:only the IOL_NLEN characters used */
  if ( iol_struct.debug_mode.value)
    aframe_mess(routine,GRR_SINFO,"stripped verb is: >%s<",verb);

  /* Now send the cleaned verb off for matching */
  return(iol_load_value_from_mod_verb(value, module, verb, module_matches,
				      verb_matches));
}
