#ifndef _ITRLEDA_H_
#define _ITRLEDA_H_

#ifdef __cplusplus
extern "C" {
#endif

#include "CVStag.itrleda"

#include "itrleda_structs.h"

/* IOL definitions
   =============== */
#include "iol.h"

typedef struct {
  IOL_MNAME  module[IOL_MLEN]; /* Module name must be first            */
  IOL_SIZE   size;             /* # of verbs must be second            */ 
  IOL_STRING version;          /* version of library                   */
  IOL_STRING mon_database;     /* database (server) for ped, noise     */
  IOL_STRING setup_database;   /* database (server) for setup          */
  IOL_STRING align_database;   /* database (server) for alignment      */
  IOL_INT    use_mon_db;       /* use mon database                     */
  IOL_INT    use_setup_db;     /* use setup database                   */
  IOL_INT    use_aling_db;     /* use alignment database               */
  IOL_INT    date;             /* date (yymmdd/yyyymmdd) for updates   */
  IOL_INT    time;             /* time (hhmmss) for updates            */
  IOL_INT    dst_on;           /* daylight savings time for updates ?  */
  IOL_INT    major;            /* major number for updates             */
  IOL_INT    minor;            /* minor number for updates             */
  IOL_INT    align_major;      /* major number for alignment tables    */
  IOL_INT    align_minor;      /* minor number for alignment tables    */
  IOL_INT    log_level;        /* LEDA log level                       */
  IOL_STRING chip_file;        /* file from which chips are read       */
  IOL_STRING chamber_file;     /* file from which chambers are read    */
  IOL_STRING fed_file;         /* file from which feds are read        */
  IOL_STRING align_file;       /* file from which alignment is read    */
  IOL_STRING chip_output;      /* output for ITR_Chip table            */
  IOL_STRING chamber_output;   /* output for ITR_Chamber table         */
  IOL_STRING fed_output;       /* output for ITR_Fed table             */
  IOL_INT    dump_chips;       /* dump ITR_Chip at start program?      */
  IOL_INT    dump_chambers;    /* dump ITR_Chamber at start program?   */
  IOL_INT    dump_feds;        /* dump ITR_Fed at start program?       */
  IOL_INT    call_LedaCheckDD; /* should itrleda call LedaCheckDD() ?  */
  IOL_INT    debug_mode;       /* debug mode ON/OFF                    */
} Itrleda_Iol_t;
extern Itrleda_Iol_t itrleda_iol;

enum {
  itrleda_CheckDD_NEVER = -1,
  itrleda_CheckDD_DBASE =  0,
  itrleda_CheckDD_ALWAYS = 1
};

/* function prototypes
   =================== */

/* program steering */
int ITR_leda_param_init();
int ITR_leda_initialize  ( int do_CheckDD ); 
int ITR_leda_startprogram( int major, int minor );

int ITR_leda_update_calib    ( int major, int minor, char *pass );
int ITR_leda_update_align    ( int major, int minor, char *pass);

/* return of time stamps */
int ITR_leda_chip_major();
int ITR_leda_chip_minor();
int ITR_leda_fed_major();
int ITR_leda_fed_minor();
int ITR_leda_chamber_major();
int ITR_leda_chamber_minor();


/* time retrieval and conversion */
void  ITR_leda_get_majorminor( int *major, int *minor );
int   datetime_to_utime( int date, int time, int dst_flag, int* utime );
char* print_date( int utime );

/* functions to 'pretty-print' chamber names */
char* ITR_leda_getChamberDescr (int id);

int ITR_leda_fetch_alignment (char *database, int align_major, int align_minor);

  int ITR_leda_Chamber_structFromTable (itrleda_Chamber_t *chambers, int maxCham, int *nCham);
  int ITR_leda_Fed_structFromTable (itrleda_Fed_t *chstat, int maxFeds, int *nFeds);
  int ITR_leda_Chip_structFromTable (itrleda_Chip_t *chips, int maxChips, int *nChips);
  int ITR_leda_Chanstat_structFromTable (itrleda_Chanstat_t *chstat, int maxChips, int *nChips);
  int ITR_leda_Ped_structFromTable (itrleda_Ped_t *ped, int maxChips, int *nChips);
  int ITR_leda_Noise_structFromTable (itrleda_Noise_t *noise, int maxChips, int *nChips);
  int ITR_leda_Align_structFromTable (itrleda_Align_t *align, int maxAlign, int *nAlign);
  int ITR_leda_nede_structFromTable (itrleda_nede_t *nede, int maxNede, int *nNede);
  int ITR_leda_nesl_structFromTable (itrleda_nesl_t *nesl, int maxNesl, int *nNesl);


  int ITR_leda_Chamber_tableFromStruct (itrleda_Chamber_t *chambers, int nChambers);
  int ITR_leda_Fed_tableFromStruct (itrleda_Fed_t *fed, int nFeds);
  int ITR_leda_Chip_tableFromStruct    (itrleda_Chip_t *chips, int nChips);
  int ITR_leda_Chanstat_tableFromStruct (itrleda_Chanstat_t *cstat, int nChips);
  int ITR_leda_Ped_tableFromStruct (itrleda_Ped_t *cstat, int nChips);
  int ITR_leda_Noise_tableFromStruct (itrleda_Noise_t *cstat, int nChips);
  int ITR_leda_Align_tableFromStruct (itrleda_Align_t *align, int nAlign);
  int ITR_leda_nede_tableFromStruct (itrleda_nede_t *nede, int nNede);
  int ITR_leda_nesl_tableFromStruct (itrleda_nesl_t *nesl, int nNesl);

  int ITR_leda_RelationsFromStruct (itrleda_fmtTable_t *itrleda_tables);


/* an "alien" ... */
int iol_scan_preferences_from_db(char* dbase, int major, int minor);


#ifdef __cplusplus
}
#endif

/* structure definitions
   ===================== */
#include "itrleda_dd.h"

#endif
