#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "rpm.h"
#include "grr.h"
#include "pub.h"
#include "cif.h"
#include "libr.h"

#include "itr_cif.h"

static int run = 1;

int main(int argc,const char **argv){
  int rc;
  int i;
  const char *cna = 0;

  for(i=1;i<argc;i++){
    if ( !strncmp(argv[i],"-cna:",5))
      cna = argv[i]+5;
  }

  /* initialize rpm */
  rc = rpm_init(argc,argv);
  if ( rc ){
    printf("%s: rpm_init complained %d\n",argv[0],rc);
    return rc;
  }
  rpm_open(RPM_REGISTER);
  if ( rc ){
    printf("%s: rpm_open complained %d\n",argv[0],rc);
    return rc;
  }

  {
    char name[200];
    rpm_stat(RPM_DOMAIN,200,name);
    rpm_stat(RPM_PROCESS_NAME,200-strlen(name),name+strlen(name));
    grr_exec(name);
  }

  /* set up publication client interface */
  rc = pub_cinit();
  if ( rc ){
    grr_error(GRR_SFATAL,"pub_cinit complained",rc);
    run = 0;
  }
  

  /* handle SIGTERM */
  trmflg((int *) &run, 0);

  /* setup CIF formatter interface */
  rc = CIF_msg_init_formatter (cna);
  if ( rc ){
    char *line = "CIF_msg_init_formatter returned error";
    printf("%s:%s %d\n",argv[0],line,rc);
    grr_error(GRR_SFATAL,line,rc);
    run = 0;
  }

  /* let user register formatted tables */

  printf("calling itr_formatter_setup\n");
  rc = itr_formatter_setup(argc,argv);
  if ( rc ){
    grr_error(GRR_SERROR,"itr_formatter_setup complaines",rc);
    run = 0;
  }

  while(run){
    static char rpm_msg_data[RPM_LMAX];
    static rpm_hdr rpm_msg_hdr;

    rc=rpm_receive(&rpm_msg_hdr,&rpm_msg_data,RPM_LMAX,BLOCK_FOREVER);
    if (  rc == GRR_CONTINUE )
      continue;
    else if ( rc ){
      grr_error(GRR_SERROR,"rpm_receive complaines",rc);
      continue;
    }
    rc = rpm_dispatch(&rpm_msg_hdr,&rpm_msg_data);
    if ( rc && rc != GRR_UNKNOWN){
      char line[256];
      sprintf(line,"rpm_dispatch complaines about msg type 0x%x",rpm_msg_hdr.type);
      grr_error(GRR_SERROR,line,rc);
    }
  }

  rpm_halt();

  return 0;
}

